package ru.dbotthepony.mc.otm.client.screen.panels

import com.mojang.blaze3d.platform.GlStateManager
import com.mojang.blaze3d.platform.InputConstants
import com.mojang.blaze3d.systems.RenderSystem
import com.mojang.blaze3d.vertex.DefaultVertexFormat
import com.mojang.blaze3d.vertex.PoseStack
import com.mojang.blaze3d.vertex.VertexFormat
import it.unimi.dsi.fastutil.chars.CharOpenHashSet
import it.unimi.dsi.fastutil.ints.Int2ObjectAVLTreeMap
import it.unimi.dsi.fastutil.ints.Int2ObjectMap
import net.minecraft.client.gui.screens.Screen
import net.minecraft.client.renderer.GameRenderer
import ru.dbotthepony.mc.otm.client.isCtrlDown
import ru.dbotthepony.mc.otm.client.isKeyDown
import ru.dbotthepony.mc.otm.client.isShiftDown
import ru.dbotthepony.mc.otm.client.minecraft
import ru.dbotthepony.mc.otm.client.render.DynamicBufferSource
import ru.dbotthepony.mc.otm.client.render.TextAlign
import ru.dbotthepony.mc.otm.client.render.drawAligned
import ru.dbotthepony.mc.otm.client.render.tesselator
import ru.dbotthepony.mc.otm.core.addAll
import ru.dbotthepony.mc.otm.core.math.RGBAColor
import ru.dbotthepony.mc.otm.milliTime

open class TextInputPanel<out S : Screen>(
	screen: S,
	parent: EditablePanel<*>?,
	x: Float = 0f,
	y: Float = 0f,
	width: Float = 10f,
	height: Float = 10f,
) : EditablePanel<S>(screen, parent, x, y, width, height) {
	private data class TextSelection(val start: Int, val end: Int) {
		val isInversed get() = start > end
		val isNotEmpty get() = start != end
		val isValid get() = start != end

		val actualStart get() = if (isInversed) this.end else this.start
		val actualEnd get() = if (isInversed) this.start else this.end

		fun sub(line: String): Pair<String, String> {
			val before = line.substring(0, actualStart.coerceIn(0, line.length))
			val selected = line.substring(actualStart.coerceIn(0, line.length), actualEnd.coerceAtMost(line.length))

			return before to selected
		}

		fun coversEntireString(value: String?): Boolean {
			if (value == null)
				return false

			return value.length <= actualEnd && actualStart <= 0
		}

		fun coversEntireLine(value: String?): Boolean {
			if (value == null)
				return true

			return value.length < actualEnd && actualStart <= 0
		}

		fun coversNewline(value: String?): Boolean {
			if (value == null)
				return actualEnd == Int.MAX_VALUE

			return value.length < actualEnd
		}
	}

	private fun putSelection(index: Int, selection: TextSelection) {
		if (selection.isValid) {
			selections[index] = selection
		} else {
			selections.remove(index)
		}
	}

	private inner class Snapshot {
		private val lines = ArrayList(this@TextInputPanel.lines) // ultra fast copy
		private val cursorLine = this@TextInputPanel.cursorLine
		private val cursorCharacter = this@TextInputPanel.cursorCharacter
		private val selections = Int2ObjectAVLTreeMap(this@TextInputPanel.selections)

		fun apply() {
			this@TextInputPanel.lines.clear()
			this@TextInputPanel.lines.addAll(lines)
			this@TextInputPanel.selections.clear()
			this@TextInputPanel.selections.putAll(selections)
			this@TextInputPanel.cursorCharacter = cursorCharacter
			this@TextInputPanel.cursorLine = cursorLine
		}

		override fun equals(other: Any?): Boolean {
			if (this === other) return true
			if (javaClass != other?.javaClass) return false

			other as TextInputPanel<*>.Snapshot

			if (lines != other.lines) return false
			if (cursorLine != other.cursorLine) return false
			if (cursorCharacter != other.cursorCharacter) return false
			if (selections != other.selections) return false

			return true
		}

		override fun hashCode(): Int {
			var result = lines.hashCode()
			result = 31 * result + cursorLine
			result = 31 * result + cursorCharacter
			return result
		}
	}

	protected var ignoreUpdates = false
	var debug = true
	var multiLine = false
	var cursorLine = 0
	var cursorCharacter = 0
	var textColor = RGBAColor.WHITE
	var cursorColor = RGBAColor.BLACK

	private var textCache: String? = null
	private val lines = ArrayList<String>()
	private val selections = Int2ObjectAVLTreeMap<TextSelection>()
	private val undo = ArrayDeque<Snapshot>()
	private val redo = ArrayDeque<Snapshot>()

	private var snapshotTimer: Long? = null

	override fun tick() {
		super.tick()

		if (snapshotTimer != null && snapshotTimer!! <= milliTime) {
			pushbackSnapshot()
			snapshotTimer = null
			redo.clear()
		}
	}

	private fun pushbackSnapshot() {
		val snapshot = Snapshot()

		if (undo.isEmpty() || undo.last() != snapshot)
			undo.addLast(snapshot)
	}

	private fun pushbackSnapshotIfNoTimer() {
		if (snapshotTimer == null)
			pushbackSnapshot()
	}

	private fun recordHistory(hard: Boolean = false) {
		if (hard) {
			pushbackSnapshot()
			snapshotTimer = null
		} else if (snapshotTimer == null) {
			snapshotTimer = milliTime + 800L
		} else if (snapshotTimer!! <= milliTime) {
			pushbackSnapshot()
			snapshotTimer = null
		}

		redo.clear()
	}

	// it gets really dirty
	// idk man :(
	private fun undo() {
		if (undo.isEmpty())
			return

		if (snapshotTimer != null || redo.isEmpty()) {
			pushbackSnapshot()
			snapshotTimer = null
		}

		val current = Snapshot()

		while (undo.isNotEmpty()) {
			val removed = undo.removeLast()
			redo.addFirst(removed)

			if (removed != current) {
				removed.apply()
				break
			}
		}
	}

	// it gets really dirty
	// idk man :(
	private fun redo() {
		if (redo.isEmpty())
			return

		val current = Snapshot()

		snapshotTimer = null

		while (redo.isNotEmpty()) {
			val removed = redo.removeFirst()
			undo.addLast(removed)

			if (current != removed) {
				removed.apply()
				break
			}
		}
	}

	operator fun get(index: Int): String? {
		return lines.getOrNull(index)
	}

	operator fun set(index: Int, value: String) {
		if (index < 0) {
			throw IndexOutOfBoundsException("negative index $index")
		}

		lines.ensureCapacity(index)

		while (lines.size <= index) {
			lines.add("")
		}

		lines[index] = value
	}

	private fun selectionIndex(index: Int) = object : Int2ObjectMap.Entry<TextSelection> {
		override fun setValue(newValue: TextSelection): TextSelection? {
			return selections.put(index, newValue)
		}

		override fun getIntKey(): Int {
			return index
		}

		override val value: TextSelection
			get() = selections[index] ?: throw NoSuchElementException()
	}

	fun insertLine(index: Int, value: String = "") {
		lines.ensureCapacity(index)

		while (lines.size < index) {
			lines.add("")
		}

		val upperLines = ArrayList<Int2ObjectMap.Entry<TextSelection>>()
		upperLines.addAll(selections.int2ObjectEntrySet().iterator(selectionIndex(index)))

		for (entry in upperLines) {
			selections.remove(entry.intKey)
		}

		for (entry in upperLines) {
			selections.put(entry.intKey + 1, entry.value)
		}

		lines.add(index, value)
	}

	fun removeLine(index: Int): Boolean {
		if (index < 0 || index >= lines.size) {
			return false
		}

		if (cursorLine == index) {
			if (index != 0)
				cursorLine--

			cursorCharacter = lines.getOrNull(cursorLine)?.length ?: 0
		}

		lines.removeAt(index)
		selections.remove(index)

		val upperLines = ArrayList<Int2ObjectMap.Entry<TextSelection>>()
		upperLines.addAll(selections.int2ObjectEntrySet().iterator(selectionIndex(index)))

		for (entry in upperLines) {
			selections.remove(entry.intKey)
		}

		for (entry in upperLines) {
			selections.put(entry.intKey - 1, entry.value)
		}

		return true
	}

	protected fun moveCursors(line: Int, character: Int, moveBy: Int) {
		@Suppress("name_shadowing")
		val moveBy = if (character + moveBy < 0) -character else moveBy

		if (cursorLine == line && cursorCharacter >= character) {
			cursorCharacter = (cursorCharacter + moveBy).coerceIn(0, this[cursorLine]?.length ?: Int.MAX_VALUE)
		}

		val selection = selections[line]

		if (selection != null) {
			val (start, end) = selection

			if (character < start) {
				putSelection(line, TextSelection(start + moveBy, end + moveBy))
			} else if (character > end && character + moveBy < end) {
				if (character + moveBy < start) {
					// selections.remove(line)
					selections[line] = TextSelection(0, 0)
				} else {
					putSelection(line, TextSelection(start, character + moveBy))
				}
			} else if (character in start .. end) {
				if (moveBy > 0 || character + moveBy <= start) {
					putSelection(line, TextSelection(start, end + moveBy))
				} else {
					putSelection(line, TextSelection(character + moveBy, end + moveBy))
				}
			}
		}
	}

	protected fun wipeSelection() {
		if (selections.isEmpty())
			return

		pushbackSnapshotIfNoTimer()
		val set = selections.int2ObjectEntrySet()

		while (set.isNotEmpty()) {
			val (lineNumber, selection) = set.last()
			selections.remove(lineNumber)
			val line = this[lineNumber] ?: continue
			if (!selection.isValid) continue // ???

			if (selection.coversEntireLine(line)) {
				removeLine(lineNumber)
			} else if (selection.coversEntireString(line)) {
				this[lineNumber] = ""
			} else if (selection.coversNewline(line)) {
				val before = line.substring(0, selection.actualStart.coerceIn(0, line.length))
				val next = this[lineNumber + 1]

				if (next == null) {
					this[lineNumber] = before
				} else {
					this[lineNumber] = before + next
					removeLine(lineNumber + 1)
				}
			} else {
				val before = line.substring(0, selection.actualStart.coerceIn(0, line.length))
				val after = line.substring(selection.actualEnd.coerceIn(0, line.length))
				this[lineNumber] = before + after
			}

			if (lineNumber < cursorLine) {
				cursorLine = lineNumber
				cursorCharacter = selection.actualStart
			} else if (lineNumber == cursorLine && cursorCharacter > selection.actualStart) {
				cursorCharacter = selection.actualStart
			}
		}
	}

	protected open fun textChanged(oldText: String, newText: String) {}

	var text: String
		get() {
			var textCache = textCache

			if (textCache == null) {
				textCache = lines.joinToString("\n")
				this.textCache = textCache
			}

			return textCache
		}
		set(value) {
			if ((textCache ?: lines.joinToString("\n")) == value)
				return

			snapshotTimer = null
			selections.clear()
			lines.clear()
			redo.clear()
			undo.clear()
			cursorLine = 0
			cursorCharacter = 0
			lines.addAll(value.split(NEWLINES))
		}

	data class CursorAdvanceResult(
		val oldLine: Int,
		val newLine: Int,
		val oldCharacter: Int,
		val newCharacter: Int,
		val couldHaveChangedLine: Boolean
	) {
		val linesChanged get() = oldLine != newLine
		val charsChanged get() = linesChanged || oldCharacter != newCharacter
		val advancedChars get() = newCharacter - oldCharacter
	}

	fun advanceCursorLeft(greedy: Boolean = false): CursorAdvanceResult {
		val oldLine = cursorLine
		val oldChar = cursorCharacter
		val line = this[cursorLine]
		var couldHaveChangedLine = false

		if (line != null && cursorCharacter > line.length) {
			cursorCharacter = line.length - 1
		}

		if (cursorCharacter > 0) {
			if (greedy && line != null) {
				cursorCharacter = greedyAdvanceLeft(line, cursorCharacter)
			} else {
				cursorCharacter--
			}
		} else if (cursorLine > 0) {
			cursorLine--
			cursorCharacter = 0
			couldHaveChangedLine = true

			@Suppress("name_shadowing")
			val line = this[cursorLine]

			if (line != null) {
				cursorCharacter = line.length
			}
		} else {
			couldHaveChangedLine = true
		}

		return CursorAdvanceResult(oldLine, cursorLine, oldChar, cursorCharacter, couldHaveChangedLine = couldHaveChangedLine)
	}

	fun advanceCursorRight(greedy: Boolean = false): CursorAdvanceResult {
		val oldLine = cursorLine
		val oldChar = cursorCharacter
		var couldHaveChangedLine = false

		val line = this[cursorLine]

		if (greedy && line != null && cursorCharacter + 1 < line.length) {
			cursorCharacter = greedyAdvanceRight(line, cursorCharacter)
		} else {
			cursorCharacter++
		}

		if (line != null && cursorCharacter > line.length) {
			couldHaveChangedLine = true

			if (lines.size <= cursorLine + 1) {
				cursorCharacter = line.length
			} else {
				cursorLine++
				cursorCharacter = 0
			}
		} else if (line == null) {
			cursorCharacter = 0
		}

		return CursorAdvanceResult(oldLine, cursorLine, oldChar, cursorCharacter, couldHaveChangedLine = couldHaveChangedLine)
	}

	private fun simulateSelectLeft(greedy: Boolean) {
		val thisLine = this[cursorLine] ?: return

		val existing = selections[cursorLine]
		val result = advanceCursorLeft(greedy)

		if (!result.linesChanged)
			putSelection(result.oldLine, TextSelection(
				existing?.start ?: result.oldCharacter,
				result.newCharacter))
		else {
			this[result.newLine]?.let {
				val existingNewline = selections[result.newLine]

				if (existingNewline == null) {
					putSelection(result.newLine, TextSelection(
						Int.MAX_VALUE,
						it.length))
				} else {
					if (!existingNewline.isInversed) {
						putSelection(result.newLine, TextSelection(
							existingNewline.start,
							result.newCharacter))
					}
				}
			}

			if (existing != null && existing.actualStart == 0 && thisLine.isEmpty() && !existing.isInversed) {
				selections.remove(result.oldLine)
			}
		}
	}

	private fun simulateSelectRight(greedy: Boolean) {
		this[cursorLine] ?: return

		val existing = selections[cursorLine]
		val result = advanceCursorRight(greedy)

		putSelection(result.oldLine, TextSelection(
			existing?.start ?: result.oldCharacter,
			if (result.couldHaveChangedLine) Int.MAX_VALUE else result.newCharacter))
	}

	override fun keyPressedInternal(key: Int, scancode: Int, mods: Int): Boolean {
		if (key == InputConstants.KEY_ESCAPE) {
			killFocus()
			return true
		}

		if (key == InputConstants.KEY_RETURN) {
			if (multiLine) {
				if (!minecraft.window.isShiftDown && !minecraft.window.isCtrlDown)
					wipeSelection()

				val line = this[cursorLine]

				if (line != null && !minecraft.window.isShiftDown) {
					if (cursorCharacter <= 0) {
						recordHistory(true)
						insertLine(cursorLine, "")

						if (!minecraft.window.isCtrlDown) {
							cursorLine++
							cursorCharacter = 0
						}
					} else if (cursorCharacter >= line.length) {
						recordHistory(true)
						insertLine(cursorLine + 1, "")

						if (!minecraft.window.isCtrlDown) {
							cursorLine++
							cursorCharacter = 0
						}
					} else {
						val before = line.substring(0, cursorCharacter)
						val after = line.substring(cursorCharacter)

						recordHistory(true)
						this[cursorLine] = before
						insertLine(cursorLine + 1, after)

						if (!minecraft.window.isCtrlDown) {
							cursorLine++
							cursorCharacter = 0
						}
					}
				} else {
					recordHistory(true)
					insertLine(cursorLine + 1)
					cursorLine++
					cursorCharacter = 0
				}

				return true
			} else {
				killFocus()
				return true
			}
		}

		if (key == InputConstants.KEY_LEFT) {
			if (!minecraft.window.isShiftDown) {
				advanceCursorLeft(minecraft.window.isCtrlDown)
				selections.clear()
			} else {
				if (this[cursorLine] == null)
					return true

				simulateSelectLeft(minecraft.window.isCtrlDown)
			}

			return true
		} else if (key == InputConstants.KEY_RIGHT) {
			if (!minecraft.window.isShiftDown) {
				advanceCursorRight(minecraft.window.isCtrlDown)
				selections.clear()
			} else {
				if (this[cursorLine] == null)
					return true

				simulateSelectRight(minecraft.window.isCtrlDown)
			}

			return true
		} else if (key == InputConstants.KEY_UP) {
			if (minecraft.window.isShiftDown) {
				if (cursorLine > 0) {
					val cursorCharacter = cursorCharacter
					val cursorLine = cursorLine

					while (this.cursorCharacter >= 0 && this.cursorLine == cursorLine) {
						simulateSelectLeft(false)
					}

					val line = this[this.cursorLine]!!

					if (cursorCharacter < line.length) {
						this.cursorCharacter = line.length

						while (this.cursorCharacter > cursorCharacter) {
							simulateSelectLeft(false)
						}
					}

					this.cursorCharacter = cursorCharacter
				}
			} else {
				if (cursorLine > 0) {
					cursorLine--
				}

				selections.clear()
			}

			return true
		} else if (key == InputConstants.KEY_DOWN) {
			if (minecraft.window.isShiftDown) {
				if (cursorLine < lines.size - 1) {
					val cursorCharacter = cursorCharacter
					val cursorLine = cursorLine

					while (this.cursorCharacter >= cursorCharacter && this.cursorLine == cursorLine) {
						simulateSelectRight(true)
					}

					val line = this[this.cursorLine]!!

					if (cursorCharacter < line.length) {
						this.cursorCharacter = 0

						while (this.cursorCharacter < cursorCharacter) {
							simulateSelectRight(false)
						}
					} else {
						val store = this.cursorLine

						for (i in 0 .. line.length) {
							simulateSelectRight(false)
						}

						this.cursorLine = store
					}

					this.cursorCharacter = cursorCharacter
				}
			} else {
				if (cursorLine < lines.size - 1) {
					cursorLine++
				}

				selections.clear()
			}

			return true
		}

		if (key == InputConstants.KEY_BACKSPACE) {
			if (selections.isNotEmpty()) {
				wipeSelection()
				return true
			}

			val line = this[cursorLine]

			if (cursorLine <= 0 && cursorCharacter <= 0) {
				return true
			} else if (cursorCharacter <= 0 || line?.length == 0) {
				if (line == null) {
					cursorCharacter = this[--cursorLine]?.length ?: 0
					recordHistory()
				} else {
					removeLine(cursorLine)
					val newLine = this[cursorLine]

					if (newLine == null) {
						this[cursorLine] = line
					} else {
						this[cursorLine] = newLine + line
					}

					recordHistory()
				}
			} else {
				if (line != null) {
					pushbackSnapshotIfNoTimer()

					// remove from very end
					if (cursorCharacter >= line.length) {
						val newLine = line.substring(0, line.length - 1)
						moveCursors(cursorLine, cursorCharacter, -1)
						this[cursorLine] = newLine
						recordHistory()
					} else {
						val newLine = line.substring(0, cursorCharacter - 1) + line.substring(cursorCharacter)
						moveCursors(cursorLine, cursorCharacter, -1)
						this[cursorLine] = newLine
						recordHistory()
					}
				} else if (cursorLine > 0) {
					cursorLine--
					recordHistory()
				}
			}

			return true
		}

		if (key == InputConstants.KEY_DELETE) {
			if (selections.isNotEmpty()) {
				wipeSelection()
				return true
			}

			if (cursorLine !in 0 until lines.size) {
				cursorLine = lines.size - 1
				return true
			}

			if (minecraft.window.isShiftDown) {
				pushbackSnapshot()
				removeLine(cursorLine)

				if (cursorLine >= lines.size)
					cursorLine = lines.size

				cursorCharacter = 0
				recordHistory(true)
				return true
			}

			val line = this[cursorLine]!!

			if (cursorCharacter >= line.length) {
				if (cursorLine + 1 == lines.size) {
					return true
				}

				pushbackSnapshotIfNoTimer()

				val bottomLine = this[cursorLine + 1]!!
				cursorCharacter = line.length
				this[cursorLine] = line + bottomLine
				removeLine(cursorLine + 1)

				recordHistory()
			} else {
				pushbackSnapshotIfNoTimer()

				val cursorCharacter = cursorCharacter
				moveCursors(cursorLine, cursorCharacter, -1)
				this[cursorLine] = line.substring(0, cursorCharacter) + line.substring(cursorCharacter + 1)

				if (cursorCharacter != 0)
					this.cursorCharacter++

				recordHistory()
			}

			return true
		}

		if (key == InputConstants.KEY_Z && minecraft.window.isCtrlDown) {
			undo()
		} else if (key == InputConstants.KEY_Y && minecraft.window.isCtrlDown) {
			redo()
		}

		if (key == InputConstants.KEY_A && minecraft.window.isCtrlDown) {
			selections.clear()

			for ((i, line) in lines.withIndex()) {
				selections[i] = TextSelection(0, Int.MAX_VALUE)
			}
		}

		return true
	}

	override fun charTypedInternal(codepoint: Char, mods: Int): Boolean {
		wipeSelection()
		var line = this[cursorLine]

		if (line == null) {
			set(cursorLine, "")
			line = ""
			cursorCharacter = 0
		}

		if (cursorCharacter >= line.length)
			line += codepoint
		else
			line = line.substring(0, cursorCharacter) + codepoint + line.substring(cursorCharacter)

		pushbackSnapshotIfNoTimer()

		set(cursorLine, line)
		moveCursors(cursorLine, cursorCharacter, 1)
		recordHistory()

		return true
	}

	override fun innerRender(stack: PoseStack, mouseX: Float, mouseY: Float, partialTick: Float) {
		var y = 0f

		for ((i, line) in lines.withIndex()) {
			val selection = selections[i]

			font.drawAligned(
				poseStack = stack,
				buffer = BUFFER,
				text = line,
				align = TextAlign.TOP_LEFT,
				x = 0f,
				y = y,
				color = textColor
			)

			if (selection != null && selection.isValid && (selection.isNotEmpty || selection.coversEntireLine(line))) {
				val (before, selected) = selection.sub(line)

				var x = 0f

				if (before.isNotEmpty()) {
					x = font.width(before).toFloat()
				}

				val width = if (selection.coversNewline(line)) this.width - x else font.width(selected).toFloat()

				RenderSystem.setShader(GameRenderer::getPositionShader)
				RenderSystem.setShaderColor(0.0f, 0.0f, 1.0f, 1.0f)
				RenderSystem.disableTexture()
				RenderSystem.enableColorLogicOp()
				RenderSystem.logicOp(GlStateManager.LogicOp.OR_REVERSE)
				RenderSystem.disableDepthTest()
				RenderSystem.defaultBlendFunc()

				val builder = tesselator.builder

				builder.begin(VertexFormat.Mode.QUADS, DefaultVertexFormat.POSITION)

				builder.vertex(stack.last().pose(), x, y + font.lineHeight + 2f, 0f).endVertex()
				builder.vertex(stack.last().pose(), x + width, y + font.lineHeight + 2f, 0f).endVertex()
				builder.vertex(stack.last().pose(), x + width, y, 0f).endVertex()
				builder.vertex(stack.last().pose(), x, y, 0f).endVertex()

				tesselator.end()

				RenderSystem.setShaderColor(1.0f, 1.0f, 1.0f, 1.0f)
				RenderSystem.disableColorLogicOp()
				RenderSystem.enableTexture()
				RenderSystem.enableDepthTest()
			}

			y += font.lineHeight + 2f
		}

		if (isFocused && milliTime % 1000L > 200L) {
			val activeLine = this[cursorLine]

			if (activeLine == null || cursorCharacter >= activeLine.length) {
				font.drawAligned(
					poseStack = stack,
					buffer = BUFFER,
					text = "_",
					align = TextAlign.TOP_LEFT,
					x = if (activeLine == null) 0f else font.width(activeLine).toFloat(),
					y = cursorLine * (font.lineHeight + 2f),
					color = cursorColor
				)
			} else {
				font.drawAligned(
					poseStack = stack,
					buffer = BUFFER,
					text = "|",
					align = TextAlign.TOP_LEFT,
					x = font.width(activeLine.substring(0, cursorCharacter)).toFloat() - 1f,
					y = cursorLine * (font.lineHeight + 2f),
					color = cursorColor
				)
			}
		}

		if (debug) {
			font.drawAligned(
				poseStack = stack,
				buffer = BUFFER,
				text = cursorLine.toString(),
				align = TextAlign.TOP_RIGHT,
				x = width,
				y = 0f,
				color = cursorColor
			)

			font.drawAligned(
				poseStack = stack,
				buffer = BUFFER,
				text = cursorCharacter.toString(),
				align = TextAlign.TOP_RIGHT,
				x = width,
				y = font.lineHeight + 2f,
				color = cursorColor
			)

			font.drawAligned(
				poseStack = stack,
				buffer = BUFFER,
				text = lines.size.toString(),
				align = TextAlign.TOP_RIGHT,
				x = width,
				y = font.lineHeight * 2 + 4f,
				color = cursorColor
			)
		}

		BUFFER.endBatch()
	}

	override fun mouseClickedInner(x: Double, y: Double, button: Int): Boolean {
		selections.clear()
		requestFocus()
		return true
	}

	private enum class CharType {
		SPACES {
			override fun contains(input: Char): Boolean {
				return _SPACES.contains(input)
			}

			override fun canJumpInto(category: CharType): Boolean {
				return true
			}
		},
		CONTROL {
			override fun contains(input: Char): Boolean {
				return _CONTROL.contains(input)
			}

			override fun canJumpInto(category: CharType): Boolean {
				return category == CONTROL
			}
		},
		OTHER {
			override fun contains(input: Char): Boolean {
				return !_SPACES.contains(input) && !_CONTROL.contains(input)
			}

			override fun canJumpInto(category: CharType): Boolean {
				return category == OTHER
			}
		};

		abstract fun contains(input: Char): Boolean
		abstract fun canJumpInto(category: CharType): Boolean

		companion object {
			private val values = values()

			operator fun get(input: Char): CharType {
				for (value in values) {
					if (value.contains(input)) {
						return value
					}
				}

				throw NoSuchElementException("for $input")
			}
		}
	}

	companion object {
		val NEWLINES = Regex("\n\r?")
		private val BUFFER = DynamicBufferSource()

		private val _SPACES = CharOpenHashSet().also {
			for (char in " \t\n\r\u0000")
				it.add(char)
		}

		private val _CONTROL = CharOpenHashSet().also {
			for (char in "!@\"'#$%^&?*()~`[]/\\;:|<>.,")
				it.add(char)
		}

		private fun greedyAdvanceLeft(input: String, position: Int): Int {
			if (position <= 1)
				return -1

			@Suppress("name_shadowing")
			var position = position
			var type = CharType[input[--position]]

			while (position > 0) {
				val newType = CharType[input[position]]

				if (type.canJumpInto(newType)) {
					position--
					type = newType
				} else {
					return position + 1
				}
			}

			return position
		}

		private fun greedyAdvanceRight(input: String, position: Int): Int {
			@Suppress("name_shadowing")
			var position = position

			if (position < 0)
				position = 0
			else if (position >= input.length)
				return position + 1

			var type = CharType[input[position]]

			while (position < input.length) {
				val newType = CharType[input[position]]

				if (type.canJumpInto(newType)) {
					position++
					type = newType
				} else {
					return position
				}
			}

			return position
		}
	}
}
